import java.lang.Math;
import java.util.StringTokenizer;


class Player
{
	public rogi2			m_rogi2;			
	public Memory			m_memory;	
	public char			m_side;
	public int			m_number;
	public String 			m_team;
	public String			m_playMode;	
	public Fuzzy			m_fuzzy;
	public int			m_rol;
	public ListPossibleActions 	m_actions;
	public PossibleAction 		reactive_decision;
	public PossibleAction 		deliverative_decision;
	public CooperativeWorldPerception m_cwp;
	public int				lastcatch;

	public static int 	SEARCH_ANGLE		= SoccerParams.VISIBLE_ANGLE;
	public int 		SEARCH_DIRECTION 	= 1;

	Functions f = new Functions();
	private Position intercept_ball_position = null;

	public Player(rogi2 r2, Memory memory, char side, String playMode, String team, int number)
	{
		m_rogi2 = r2;
		m_memory = memory;
		m_side = side;
		m_number = number;
		m_team = team;
		m_fuzzy = new Fuzzy(m_memory, m_side, m_number, m_team);
		m_playMode = playMode;
		m_rol = m_number-1;
		m_actions = new ListPossibleActions();
		reactive_decision = null;
		deliverative_decision = null;
		m_cwp = new CooperativeWorldPerception(this);
		lastcatch = -10;
	}
	

	/********************************************************************************/
	/*************************   DECISION SYSTEM   **********************************/
	/********************************************************************************/
	
	public void TakeReactiveDecision(Memory memory, String playMode)
	{
		// this method is implemented in the subclass because depends of the player behavior
	}

	public void SendReactiveDecision(BallInfo ball)
	{
		HearInfo info = new HearInfo(0, reactive_decision.time, m_number, m_memory.myX, m_memory.myY, 
				m_memory.myDir,	m_memory.m_velX, m_memory.m_velY, m_memory.m_prediction, 
				reactive_decision.action, reactive_decision.certainty, reactive_decision.dir, 
				m_rol, m_memory.m_stamina, m_memory.m_effort, m_memory.m_recover);

		m_rogi2.Say(info.message, 0);
	}

	public void TakeDeliberativeDecision()
	{
		ListPossibleActions o_actions = new ListPossibleActions(m_memory.o_actions);
		ListPossibleActions m_deliberatives = m_cwp.DeliberativeDecision(o_actions, m_actions);

		deliverative_decision = m_deliberatives.MoreCertainty(m_number);

		ExecuteDeliberativeDecision();
	}

	public void ExecuteDeliberativeDecision()
	{
		BallInfo ball = m_memory.ball;
		if (deliverative_decision.action != Actions.SEARCH_BALL) ball.lost = false;

		if (deliverative_decision.action == Actions.KICK_OFF)
		{
//System.out.println("KICK_KOFF "+deliverative_decision.certainty);
			KickOff(ball);
		}
		else if (deliverative_decision.action == Actions.MOVE)
		{
//System.out.println("MOVE "+deliverative_decision.certainty);
			if (m_playMode.compareTo("free_kick_"+m_side) == 0) GoalieMove();
			else StartPosition();
		}
		else if (deliverative_decision.action == Actions.NOP)
		{
//System.out.println("NOP "+deliverative_decision.certainty);
			TurnTo(ball);
		}
		else if (deliverative_decision.action == Actions.SEARCH_BALL)
		{
//System.out.println("SEARCH_BALL "+deliverative_decision.certainty);
			Search(ball);
		}
		else if (deliverative_decision.action == Actions.ZONE)
		{
//System.out.println("ZONE "+deliverative_decision.certainty);
			Zone(ball);
		}
		else if (deliverative_decision.action == Actions.CLEARING)
		{
//System.out.println("CLEARING "+deliverative_decision.certainty);
			Clearing(ball);
		}
		else if (deliverative_decision.action == Actions.KICK_GOAL)
		{
//System.out.println("KICK_GOAL "+deliverative_decision.certainty);
			KickGoal(ball);
		}
		else if (deliverative_decision.action == Actions.CONTROL_BALL)
		{
//System.out.println("CONTROL_BALL "+deliverative_decision.certainty);
			ControlBall(f.ControlDirection(m_rol, m_side, m_memory.myX, m_memory.myY, m_memory.myDir), ball);
		}
		else if (deliverative_decision.action == Actions.DRIBBLE)
		{
//System.out.println("DRIBBLE "+deliverative_decision.certainty);
			Dribble(ball);
		}
		else if (deliverative_decision.action == Actions.INTERCEPT_BALL)
		{
//System.out.println("INTERCEPT_BALL "+deliverative_decision.certainty);
			InterceptBall(ball);
		}
		else if (deliverative_decision.action == Actions.PASS)
		{
//System.out.println("PASS "+deliverative_decision.certainty);
			PassTo(FindPass(ball), ball);
		}
		else if (deliverative_decision.action == Actions.CATCH_BALL)
		{
//System.out.println("CATCH_BALL "+deliverative_decision.certainty);
			lastcatch = m_memory.m_time;
			CatchBall(ball);
		}
		else if (deliverative_decision.action == Actions.CENTER)
		{
//System.out.println("CENTER "+deliverative_decision.certainty);
			Center(ball);
		}
	}

	/********************************************************************************/
	/*************************   BASIC ACTIONS   ************************************/
	/********************************************************************************/
	
	public void Move(int x, int y)
	{
		m_memory.myX = x;
		m_memory.myY = y;

		m_rogi2.move(x, y);
	}

	public void Turn(float dir)
	{
// the player has bad movements
//			float moment = (float)(dir * (1.0 + SoccerParams.INERTIA_MOMENT*m_memory.m_speed));
//			if (moment > SoccerParams.MAX_MOMENT) moment = SoccerParams.MAX_MOMENT;
//			else if (moment < SoccerParams.MIN_MOMENT) moment = SoccerParams.MIN_MOMENT;
//			m_memory.TurnmyDir(moment);
//			m_rogi2.turn(moment);

		m_memory.TurnmyDir(dir);
		m_rogi2.turn(dir);
	}

	public void Dash(int power)
	{
		if (power>SoccerParams.MAX_POWER) power=SoccerParams.MAX_POWER;
		else if (power<SoccerParams.MIN_POWER) power=SoccerParams.MIN_POWER;

		// #########  stamina save  ############

		BallInfo ball = m_memory.ball;
		char o_side = 'l';
		if (m_side == 'l') o_side = 'r';
		if ((m_playMode.compareTo("kick_off_"+m_side) == 0)
		  ||((m_playMode.compareTo("corner_kick_"+o_side) != 0)
			  &&(m_playMode.compareTo("free_kick_"+o_side) != 0)
			  &&(m_playMode.compareTo("goal_kick_"+o_side) != 0)
			  &&(m_playMode.compareTo("kick_in_"+o_side) != 0)
			  &&(deliverative_decision.action != Actions.INTERCEPT_BALL)
			  &&(ball.m_distance > 20))
		  ||((m_playMode.compareTo("kick_in_"+m_side) == 0)&&(deliverative_decision.action == Actions.INTERCEPT_BALL))
		  ||((m_playMode.compareTo("free_kick_"+m_side) == 0)&&(deliverative_decision.action == Actions.INTERCEPT_BALL))
		  ||((m_playMode.compareTo("goal_kick_"+m_side) == 0)&&(deliverative_decision.action == Actions.INTERCEPT_BALL))
		  ||((m_playMode.compareTo("corner_kick_"+m_side) == 0)&&(deliverative_decision.action == Actions.INTERCEPT_BALL))) 
			power = (int)SoccerParams.STAMINA_INC_MAX;

		// #####################################

		m_memory.SetmyVel(power);

		m_rogi2.dash(power);
	}
	
	public void Kick(int power, float dir)
	{
		if (power>SoccerParams.MAX_POWER) power=SoccerParams.MAX_POWER;
		else if (power<SoccerParams.MIN_POWER) power=SoccerParams.MIN_POWER;

		m_memory.SetBallVel(power, dir);

		m_rogi2.kick(power, dir);
	}

	public void Catch(float dir)
	{
		m_rogi2.Catch(dir);
	}


	/********************************************************************************/
	/**********************    ACTIONS WITHOUT THE BALL   ***************************/
	/********************************************************************************/

	public void StartPosition()
	{
		Move(f.FirstXPosition(m_rol),f.FirstYPosition(m_rol));
	}

	public void GoalieMove()
	{
		if ( m_memory.ball.m_position.Y <= 0) Move(-37, -20);
		else Move(-37, 20);
	}

	public void KickOff(BallInfo ball)
	{
		if (ball.m_distance > SoccerParams.KICKABLE_AREA) GoTo(ball.m_position, ball);
		else 
		{
			PlayerInfo player = FindPass(ball);
			if (player != null) PassTo(player, ball);
			else ControlBall(f.ControlDirection(m_rol, m_side, m_memory.myX, m_memory.myY, m_memory.myDir), ball);
		}
	}

	public void Zone(BallInfo ball)
	{
		float x = f.XZone(m_rol, ball);
		float y = f.YZone(m_rol, ball);

		if ((x == -1)||(y == -1))
		{
			if (m_rol == Rols.PORTER) ProtectGoal(ball, 2);
			else if (m_rol == Rols.CENTRAL) ProtectGoal(ball, 16);
			else if ((m_rol == Rols.MIG_DRET)
				||(m_rol == Rols.MIG_ESQUERRA)
				||(m_rol == Rols.MITJA_PUNTA)) FollowBall(ball);
			else AtackSituation(ball);
		}
		else
		{
			Position p = new Position(m_side, x, y, m_memory.myX, m_memory.myY, m_memory.myDir);
			GoTo(p, ball);
		}
	}

	public void AtackSituation(BallInfo ball)
	{
		PlayerInfo player = TheirLastDefensor();
		float x=0;

		if (player != null)
		{
			x=player.m_position.X;
			if (x<0) x=0;
		}

		if (ball.m_position.X > x) x=ball.m_position.X;

		Position p = new Position(m_side, x-2, f.FirstYPosition(m_rol), m_memory.myX, m_memory.myY, m_memory.myDir);

		if (m_fuzzy.InsideZone(p.dist, 1, 3) < 0.7) GoTo(p, ball);
		else TurnTo(ball);
	}

	public Position ProtectGoalPosition(Position ball, float distance)
	{
		Position aux = new Position(m_side, ball.X, ball.Y, -(float)SoccerParams.FIELD_WIDTH/2, 0, 0);
		Position result = new Position(aux.dir, distance, m_side, -(float)SoccerParams.FIELD_WIDTH/2, 0, 0);

		return result;
	}
	public void ProtectGoal(BallInfo ball, float distance)
	{
		Position pgp = ProtectGoalPosition(ball.m_position , distance);
		Position p = new Position(m_side, pgp.X, pgp.Y, m_memory.myX, m_memory.myY, m_memory.myDir);

		if (m_fuzzy.InsideZone(p.dist, 1, 3) < 0.7) GoTo(p, ball);
		else TurnTo(ball);
	}

	public void FollowBall(BallInfo ball)
	{
		Position p = new Position(m_side, ball.m_position.X-2, f.FirstYPosition(m_rol), m_memory.myX, m_memory.myY, m_memory.myDir);

		if (m_fuzzy.InsideZone(p.dist, 1, 3) < 0.7) GoTo(p, ball);
		else TurnTo(ball);
	}

	public void DashDist(float dist)
	{
		int power = (int)((dist-m_memory.m_speed)/(SoccerParams.DASH_POWER_RATE));
		if (power > 0) Dash(power);
	}

	public void AvoidPlayer(PlayerInfo player)
	{
		if (player != null)
		{
			Position aux = new Position(player.m_position);

			float x = aux.dist*(float)Math.cos(f.radiants(aux.dir));
			float y = aux.dist*(float)Math.sin(f.radiants(aux.dir));

			if (player.m_direction < 0) y = (float)(3*SoccerParams.CLOSE_MARGIN) + y;
			else y = -((float)(3*SoccerParams.CLOSE_MARGIN) - y);

			float avoid_angle = 0;
			if (x != 0) avoid_angle = f.degrees((float)Math.atan(y/x));

			if (((player.m_direction<0)&&(avoid_angle>0)) || ((player.m_direction>0)&&(avoid_angle<0))) Turn(avoid_angle);
			else
			{
				float dist = (float)Math.sqrt(x*x + y*y);
				DashDist(dist);
			}
		}
	}

	public void GoTo(Position pos, BallInfo ball)
	{
		if (!FreeSector(pos.dir, (float)(4*SoccerParams.CLOSE_MARGIN), 120))
		{
			AvoidPlayer(NearOppSector(pos.dir, (float)(4*SoccerParams.CLOSE_MARGIN) , 120));
		}
		else if (!pos.GoodDirection()) Turn(pos.dir); 
		else DashDist(pos.dist);
	}

	public void TurnTo(ObjectInfo object)
	{
		Turn(f.StandarizeAngle(object.m_direction));
	}

	public void Search()
	{
		Turn(SEARCH_DIRECTION*SEARCH_ANGLE);

		m_memory.wait=true;	// wait a see information
	}

	public void Search(BallInfo ball)
	{
		if (ball.lost == true) Turn(SEARCH_DIRECTION*SEARCH_ANGLE);
		else
		{
			if (Math.abs(ball.m_direction) < (SoccerParams.VISIBLE_ANGLE/2))
			{
				Turn(SEARCH_DIRECTION*SEARCH_ANGLE);
				ball.lost=true;
			}
			else 
			{
				if (ball.m_direction<0) SEARCH_DIRECTION=-1;
				else SEARCH_DIRECTION=1;
	
				Turn(ball.m_direction);
			}
		}

		m_memory.wait=true;	// wait a see information
	}

	public Position InterceptBallPosition(BallInfo ball)
	{
		BallInfo future_ball = new BallInfo(ball);
		Position m_f_pos = new Position(0, 0, m_side, m_memory.myX, m_memory.myY, m_memory.myDir);
		float dist = m_f_pos.Distance(future_ball.m_position);

		int c = 0;
		while (dist > SoccerParams.KICKABLE_AREA)
		{
			c++;

			future_ball = PredictBall(future_ball, 1);	
			m_f_pos = PlayerPredictedPosition(m_memory.myX, m_memory.myY, m_memory.m_velX, m_memory.m_velY, 
							  m_memory.myDir, m_memory.m_stamina, m_memory.m_effort, 
							  m_memory.m_recover, future_ball.m_direction, c, 100);

			dist = m_f_pos.Distance(future_ball.m_position);

			if (c > 50) return future_ball.m_position;
		}
		return future_ball.m_position;
	}

	public Position GoalieInterceptBallPosition(BallInfo ball)
	{
		BallInfo future_ball = new BallInfo(ball);
		float y = future_ball.m_position.Y;
		if (y < -SoccerParams.GOAL_WIDTH/2) y = -(float)SoccerParams.GOAL_WIDTH/2;
		else if (y > SoccerParams.GOAL_WIDTH/2) y = (float)SoccerParams.GOAL_WIDTH/2;
		float x = -53;
		Position aux = new Position(m_side, x, y, m_memory.myX, m_memory.myY, m_memory.myDir);
		float dist = aux.Distance(future_ball.m_position);

		int c = 0;
		while (dist > SoccerParams.KICKABLE_AREA)
		{
			c++;

			future_ball = PredictBall(future_ball, 1);

			if (future_ball.m_distance < SoccerParams.KICKABLE_AREA/2)
			{
				Position aux2 = new Position(0, 0, m_side, m_memory.myX, m_memory.myY, m_memory.myDir);
				return aux2;
			}

			y = future_ball.m_position.Y;
			if (y < -SoccerParams.GOAL_WIDTH/2) y = -(float)SoccerParams.GOAL_WIDTH/2;
			else if (y > SoccerParams.GOAL_WIDTH/2) y = (float)SoccerParams.GOAL_WIDTH/2;

			aux = new Position(m_side, x, y, m_memory.myX, m_memory.myY, m_memory.myDir);
			dist = aux.Distance(future_ball.m_position);

			if (c > 30) 
			{
				Position aux2 = new Position(m_side, -50, aux.Y, m_memory.myX, m_memory.myY, m_memory.myDir);
				return aux2;
			}
		}

		Position aux2 = new Position(m_side, -50, aux.Y, m_memory.myX, m_memory.myY, m_memory.myDir);
		return aux2;
	}

	public void InterceptBall(BallInfo ball)
	{
		if ((m_rol == Rols.PORTER)
		  &&(m_playMode.compareTo("free_kick_"+m_side)!=0)
		  &&(m_playMode.compareTo("goal_kick_"+m_side)!=0))
		{
			if (ball.m_position.X < -48) GoTo(ball.m_position, ball);
			else
			{
				Position p = GoalieInterceptBallPosition(ball);
				if (m_fuzzy.InsideZone(p.dist, 0, 2) < 0.5) GoTo(p, ball);
				else TurnTo(ball);
			}
		}
		else
		{
			Position p = InterceptBallPosition(ball);
			Position aux = null;
			if (intercept_ball_position == null) aux = p;
			else
			{
				float d = p.Distance(intercept_ball_position);

				if (m_fuzzy.InsideZone(d, (float)0, 2) < 0.5) aux = p;
				else aux = intercept_ball_position;
			}

			intercept_ball_position = new Position(m_side, aux.X, aux.Y, m_memory.myX, m_memory.myY, m_memory.myDir);

			GoTo(intercept_ball_position, ball);
		}
	}


	/********************************************************************************/
	/***********************    ACTIONS WITH THE BALL   *****************************/
	/********************************************************************************/

	public void CatchBall(BallInfo ball)
	{
		Catch(ball.m_direction);
	}

	public Position CorrectWithMyVel(float dir, float dist)
	{
		Position aux = new Position(dir, dist, m_side, m_memory.myX, m_memory.myY, m_memory.myDir);		
		float newx = aux.X + m_memory.m_velX;
		float newy = aux.Y + m_memory.m_velY;
		Position aux2 = new Position(m_side, newx, newy, m_memory.myX, m_memory.myY, m_memory.myDir);	

		return aux2;
	}

	public void TurnWithBall(BallInfo ball, float enddir)
	{
		float dir;
		float dist;

		int turn_dir = 1;
		float turn1 = enddir - ball.m_direction;
		float turn2 = ball.m_direction - enddir;
		if (turn1 < 0) turn1 += 360;
		if (turn2 < 0) turn2 += 360;
		if (turn1 > turn2) turn_dir = -1;

		if (ball.m_distance < SoccerParams.CLOSE_MARGIN)	// the ball is too close, rotate 90 degrees
		{
			dir = ball.m_direction + 180 - 90*turn_dir;
			dist = (float)Math.sqrt(Math.abs(SoccerParams.CONTROL_DIST*SoccerParams.CONTROL_DIST-ball.m_distance*ball.m_distance));
		}
		else
		{
			dir = ball.m_direction + 180 - turn_dir*f.degrees((float)Math.asin(SoccerParams.CLOSE_MARGIN/ball.m_distance));
			dist = (float)Math.sqrt(ball.m_distance*ball.m_distance-SoccerParams.CLOSE_MARGIN*SoccerParams.CLOSE_MARGIN);
			dist += (float)Math.sqrt(SoccerParams.CONTROL_DIST*SoccerParams.CONTROL_DIST-SoccerParams.CLOSE_MARGIN*SoccerParams.CLOSE_MARGIN);
		}

		Position corrected_kick = CorrectWithMyVel(dir, dist);
		KickDist(corrected_kick.dir, corrected_kick.dist, ball);
	}

	public void KickDist(float dir, float dist, BallInfo ball)
	{
		float ball_vel = (float)Math.sqrt(ball.m_velX*ball.m_velX+ball.m_velY*ball.m_velY);

		Kick((int)((dist-ball_vel)/ball.KickRate()), dir);
	}

	public void MaxKick(float dir, BallInfo ball)
	{
		KickBall(dir, 9999, ball);
	}

	public void KickBall(float dir, float dist, BallInfo ball)
	{
		if (DirectKick(dir, dist, ball)) KickDist(dir, dist, ball);
		else TurnWithBall(ball, dir);
	}

	public void ControlBall(float dir, BallInfo ball)
	{
		float dist = 2*(float)SoccerParams.CONTROL_DIST-ball.m_distance;
		KickDist(dir, dist, ball);
	}

	public boolean DirectKick(float dir, float dist, BallInfo ball)
	{
		Position aux = new Position(dir, dist, m_side, m_memory.myX, m_memory.myY, m_memory.myDir);
		float d1 = aux.dist;

		float dx = aux.X - ball.m_position.X;
		float dy = aux.Y - ball.m_position.Y;
		float d2 = (float)Math.sqrt(dx*dx + dy*dy);

		float d3 = ball.m_distance;

		float a = f.degrees((float)(Math.acos((d1*d1 - d2*d2 - d3*d3) / (-2*d3*d2))));
		if (Math.abs(a) > 90) return true;

		float d = (float)(Math.sin(f.radiants(a))*d3);

		return (Math.abs(d) > SoccerParams.CLOSE_MARGIN);
	}

	public void KickGoal(BallInfo ball)
	{
		float condition1;
		float condition2;
		float bestkickdir = 0;
		float bestcertainty = 0;
		float kick = 0;
		Position aux = null;

		for (float c = 0; c <= SoccerParams.GOAL_WIDTH/2; c+= (float)0.5)
		{
			condition1 = 0;
			condition2 = 0;

			if (ball.m_position.Y > 0) aux = new Position(m_side, (float)(SoccerParams.FIELD_WIDTH/2), c, m_memory.myX, m_memory.myY, m_memory.myDir);
			else aux = new Position(m_side, (float)(SoccerParams.FIELD_WIDTH/2), -c, m_memory.myX, m_memory.myY, m_memory.myDir);

			condition1 = m_fuzzy.AccessibleKick(aux.dir, aux.dist, ball);
			condition2 = f.not(m_fuzzy.BusySector(aux.dir, aux.dist, SoccerParams.KICK_VIEW_MOMENT));
			kick = f.and(condition1, condition2);

			if (kick > bestcertainty)
			{
				bestcertainty = kick;
				bestkickdir = aux.dir;
			}

			if (ball.m_position.Y > 0) aux = new Position(m_side, (float)(SoccerParams.FIELD_WIDTH/2), -c, m_memory.myX, m_memory.myY, m_memory.myDir);
			else aux = new Position(m_side, (float)(SoccerParams.FIELD_WIDTH/2), c, m_memory.myX, m_memory.myY, m_memory.myDir);

			condition1 = m_fuzzy.AccessibleKick(aux.dir, aux.dist, ball);
			condition2 = f.not(m_fuzzy.BusySector(aux.dir, aux.dist, SoccerParams.KICK_VIEW_MOMENT));
			kick = f.and(condition1, condition2);

			if (kick > bestcertainty)
			{
				bestcertainty = kick;
				bestkickdir = aux.dir;
			}
		}

		MaxKick(bestkickdir, ball);
	}

	public void Dribble(BallInfo ball)
	{
		float enddir = f.ControlDirection(m_rol, m_side, m_memory.myX, m_memory.myY, m_memory.myDir);
		Dribble(NearOppSector(enddir, (float)(5*SoccerParams.KICKABLE_AREA) , (float)(SoccerParams.CONTROL_VIEW_MOMENT)), enddir, ball);
	}

	public void Dribble(PlayerInfo player, float enddir, BallInfo ball)
	{
		if (player != null)
		{
			Position aux = new Position(player.m_position);

			float x = aux.dist*(float)Math.cos(f.radiants(aux.dir));
			float y = aux.dist*(float)Math.sin(f.radiants(aux.dir));

			if (player.m_direction < 0) y = (float)(3*SoccerParams.CLOSE_MARGIN) + y;
			else y = -((float)(3*SoccerParams.CLOSE_MARGIN) - y);

			float dribble_angle = 0;
			if (x != 0) dribble_angle = f.degrees((float)Math.atan(y/x));

			if (((player.m_direction<0)&&(dribble_angle>0)) || ((player.m_direction>0)&&(dribble_angle<0))) ControlBall(dribble_angle, ball);
			else ControlBall(enddir, ball);
		}
	}

	public PlayerInfo FindPass(BallInfo ball)
	{
		float certainty = -1;
		float pass;
		float condition1 = 0;
		float condition2 = 0;
		float condition3 = 0;
		float condition4 = 0;
		PlayerInfo player = null;
		PlayerInfo result = null;

		for (int priority = 1; priority <= 6; priority++)
		{
			player = m_memory.getPlayer(m_team, f.PassPriority(m_number, priority));

			if (player != null)
			{
				condition1 = m_fuzzy.GoodPrediction(player);
				condition2 = m_fuzzy.AccessibleKick(player.m_direction, player.m_distance, ball);
				condition3 = m_fuzzy.BusySector(player.m_direction, player.m_distance-1, SoccerParams.PASS_VIEW_MOMENT);
				condition4 = m_fuzzy.GoodPass(player);

				pass = f.and(condition1, f.and(condition2, f.and(f.not(condition3), condition4)));

				if (pass > certainty)
				{
					certainty = pass;
					result = player;
				}
			}
		}

		return result;
	}

	public void PassTo(PlayerInfo player, BallInfo ball)
	{
		KickBall(player.m_direction, player.m_distance, ball);
	}

	public void Clearing(BallInfo ball)
	{
		float bestkickdir = 0;
		float bestcertainty = 0;
		float kick = 0;
		Position aux = null;

		float c = (float)SoccerParams.FIELD_HEIGHT/2;
		c -= 2; 
		while ((c >= 0)&&(bestcertainty != 1))
		{
			if (ball.m_position.Y > 0) aux = new Position(m_side, (float)(SoccerParams.FIELD_WIDTH/2), c, m_memory.myX, m_memory.myY, m_memory.myDir);
			else aux = new Position(m_side, (float)(SoccerParams.FIELD_WIDTH/2), -c, m_memory.myX, m_memory.myY, m_memory.myDir);

			kick = f.not(m_fuzzy.BusySector(aux.dir, 30, SoccerParams.KICK_VIEW_MOMENT));

			if (kick > bestcertainty)
			{
				bestcertainty = kick;
				bestkickdir = aux.dir;
			}

			c--;
		}

		MaxKick(bestkickdir, ball);
	}

	public void Center(BallInfo ball)
	{
		Position aux;
		float kick;
		float bestkickdir = -90;
		float bestcertainty = 0;
		float XMARK = (float)(8*SoccerParams.FIELD_WIDTH/21);
		int c = 0;

		while ((c <= SoccerParams.FIELD_WIDTH/2 - XMARK)&&(bestcertainty != 1))
		{
			aux = new Position(m_side, XMARK+c, 0, m_memory.myX, m_memory.myY, m_memory.myDir);

			kick = f.not(m_fuzzy.BusySector(aux.dir, aux.dist, SoccerParams.KICK_VIEW_MOMENT));

			if (kick > bestcertainty)
			{
				bestcertainty = kick;
				bestkickdir = aux.dir;
			}

			c++;
		}

		MaxKick(bestkickdir, ball);
	}


	/********************************************************************************/
	/**************************      PREDICTIONS       ******************************/
	/********************************************************************************/


	public BallInfo PredictBall(BallInfo ball, int cicles)
	{
		BallInfo f_ball = new BallInfo(ball);

		for (int c=0; c < cicles; c++)
		{
			Position aux = new Position(m_side, f_ball.m_position.X+f_ball.m_velX, f_ball.m_position.Y+f_ball.m_velY, m_memory.myX, m_memory.myY, m_memory.myDir);

			f_ball.m_position=aux;
			f_ball.m_direction = aux.dir;
			f_ball.m_distance = aux.dist;
			f_ball.CalculateFlags();				

			f_ball.m_velX *= SoccerParams.BALL_DECAY;
			f_ball.m_velY *= SoccerParams.BALL_DECAY;
		}

		return f_ball;
	}

	public Position PlayerPredictedPosition(float x, float y, float vx, float vy, float m_Dir, float predicted_stamina,
				         float predicted_effort, float predicted_recovery, float dir, int steps, 
				         float dash_power)
	{
		float myang = m_Dir + dir;
		float effective_power = dash_power;

		for (int i=1; i < steps; i++) 
		{
			if (dash_power < 0) 
			{
				if (-2*dash_power > predicted_stamina) effective_power = -predicted_stamina;
				else effective_power = dash_power;
			}
			else effective_power = (float)Math.min(dash_power, predicted_stamina);
    
			effective_power *= predicted_effort;
    			effective_power *= SoccerParams.DASH_POWER_RATE;
			vx += (float)(effective_power*Math.cos(f.radiants(myang)));
			vy += (float)(effective_power*Math.sin(f.radiants(myang)));
			float v = (float)Math.sqrt(vx*vx + vy*vy);
    
    			if (v > SoccerParams.PLAYER_SPEED_MAX) 
			{
				v = (float)SoccerParams.PLAYER_SPEED_MAX;
				vx =  v*(float)Math.cos(f.radiants(myang));
				vy =  v*(float)Math.sin(f.radiants(myang));
			}
    
			x += vx;
			y += vy;

			vx *= SoccerParams.PLAYER_DECAY;
			vy *= SoccerParams.PLAYER_DECAY;

			if (dash_power < 0) predicted_stamina = (float)(Math.max(predicted_stamina+2*dash_power, 0));
			else predicted_stamina = (float)(Math.max(predicted_stamina-dash_power, 0));

			if (predicted_stamina <= SoccerParams.RECOVER_DEC_THR*SoccerParams.STAMINA_MAX)
			{
				predicted_recovery = (float)(Math.max(SoccerParams.RECOVER_MIN, predicted_recovery-SoccerParams.RECOVER_DEC));
			}

			if (predicted_stamina <= SoccerParams.EFFORT_DEC_THR*SoccerParams.STAMINA_MAX)
			{
				predicted_effort = (float)(Math.max(SoccerParams.EFFORT_MIN, predicted_effort-SoccerParams.EFFORT_DEC));
			}
			else if (predicted_stamina >= SoccerParams.EFFORT_INC_THR*SoccerParams.STAMINA_MAX)
			{
				predicted_effort = (float)(Math.min(1.0, predicted_effort+SoccerParams.EFFORT_INC));
			}

			predicted_stamina = (float)(Math.min(SoccerParams.STAMINA_MAX, predicted_stamina+(predicted_recovery*SoccerParams.STAMINA_INC_MAX)));
  		}

		Position position = new Position(m_side, x, y, m_memory.myX, m_memory.myY, m_memory.myDir);

  		return position;
  	}

	public int MyCiclesToInterceptBall(BallInfo ball)
	{
		BallInfo future_ball = new BallInfo(ball);
		Position m_f_pos = new Position(0, 0, m_side, m_memory.myX, m_memory.myY, m_memory.myDir);
		float dist = m_f_pos.Distance(future_ball.m_position);

		int c = 0;
		while (dist > SoccerParams.KICKABLE_AREA)
		{
			c++;

			future_ball = PredictBall(future_ball, 1);	
			m_f_pos = PlayerPredictedPosition(m_memory.myX, m_memory.myY, m_memory.m_velX, m_memory.m_velY, 
							  m_memory.myDir, m_memory.m_stamina, m_memory.m_effort, 
							  m_memory.m_recover, future_ball.m_direction, c, 100);

			dist = m_f_pos.Distance(future_ball.m_position);

			if (c > 50) return 100;
		}

		return c;
	}

	public int PlayerCiclesToInterceptBall(PlayerInfo player, BallInfo ball)
	{
		BallInfo future_ball = new BallInfo(ball);
		Position future_player = new Position(player.m_position);
		float dist = future_player.Distance(future_ball.m_position);

		int c = 0;

		while (dist > SoccerParams.KICKABLE_AREA)
		{
			c++;

			future_ball = PredictBall(future_ball, 1);	
			future_player = PlayerPredictedPosition(player.m_position.X, player.m_position.Y, 
							player.m_velX, player.m_velY, player.FaceDir+m_memory.myDir, 
							player.m_stamina, player.m_effort, player.m_recover, 
							future_ball.m_direction, c, 100);

			dist = future_player.Distance(future_ball.m_position);

			if (c > 50) return 100;
		}

		return c;
	}

	/********************************************************************************/
	/********************************   MISC    *************************************/
	/********************************************************************************/


	public boolean FreeSector(float dir, float dist, float moment)
	{
		boolean free=true;

		int c=0; 
		while ((c < m_memory.m_players.size())&&(free))
		{
			PlayerInfo player = (PlayerInfo)m_memory.m_players.elementAt(c);

			if ((player.m_direction < f.StandarizeAngle(dir+moment/2))&&(player.m_direction > f.StandarizeAngle(dir-moment/2)))
			{
				if (player.m_distance <= dist) free=false;
			}
			c++;
		}
		c=0; 
		while ((c < m_memory.o_players.size())&&(free))
		{
			PlayerInfo player = (PlayerInfo)m_memory.o_players.elementAt(c);

			if ((player.m_direction < f.StandarizeAngle(dir+moment/2))&&(player.m_direction > f.StandarizeAngle(dir-moment/2)))
			{
				if (player.m_distance <= dist) free=false;
			}
			c++;
		}

		return free;
	}

	public boolean FreeZone(float dir, float dist, float height)
	{
		boolean free=true;

		int c=0; 
		while ((c < m_memory.m_players.size())&&(free))
		{
			PlayerInfo player = (PlayerInfo)m_memory.m_players.elementAt(c);

			float h = (float)(player.m_distance*Math.sin(f.radiants(player.m_direction)));

			if ((player.m_distance <= dist)&&(h <= height)) free=false;

			c++;
		}
		c=0; 
		while ((c < m_memory.o_players.size())&&(free))
		{
			PlayerInfo player = (PlayerInfo)m_memory.o_players.elementAt(c);

			float h = (float)(player.m_distance*Math.sin(f.radiants(player.m_direction)));

			if ((player.m_distance <= dist)&&(h <= height)) free=false;

			c++;
		}

		return free;
	}

	public PlayerInfo NearOppZone(float dir, float dist, float height)
	{
		PlayerInfo near = null;

		for (int c=0; c < m_memory.m_players.size(); c++)
		{
			PlayerInfo player = (PlayerInfo)m_memory.m_players.elementAt(c);

			float h = (float)(player.m_distance*Math.sin(f.radiants(player.m_direction)));

			if ((player.m_distance <= dist)&&(h <= height)) 
			{
				near = player;
				dist= player.m_distance;
			}
		}
		for (int c=0; c < m_memory.o_players.size(); c++)
		{
			PlayerInfo player = (PlayerInfo)m_memory.o_players.elementAt(c);

			float h = (float)(player.m_distance*Math.sin(f.radiants(player.m_direction)));

			if ((player.m_distance <= dist)&&(h <= height)) 
			{
				near = player;
				dist= player.m_distance;
			}
		}

		return near;
	}

	public PlayerInfo NearOppSector(float dir, float dist, float moment)
	{
		PlayerInfo near = null;

		for (int c=0; c < m_memory.m_players.size(); c++)
		{
			PlayerInfo player = (PlayerInfo)m_memory.m_players.elementAt(c);

			if ((player.m_direction < f.StandarizeAngle(dir+moment/2))&&(player.m_direction > f.StandarizeAngle(dir-moment/2)))
			{
				if (player.m_distance <= dist)
				{
					near = player;
					dist= player.m_distance;
				}
			}
		}
		for (int c=0; c < m_memory.o_players.size(); c++)
		{
			PlayerInfo player = (PlayerInfo)m_memory.o_players.elementAt(c);

			if ((player.m_direction < f.StandarizeAngle(dir+moment/2))&&(player.m_direction > f.StandarizeAngle(dir-moment/2)))
			{
				if (player.m_distance <= dist)
				{
					near = player;
					dist= player.m_distance;
				}
			}
		}

		return near;
	}

	public PlayerInfo LocalNearBall(BallInfo ball)
	{
		PlayerInfo local = null;
		float d = 99;
		
		int size = m_memory.m_players.size();
		for (int c=0; c < size; c++)
		{
			PlayerInfo player = (PlayerInfo)m_memory.m_players.elementAt(c);

			Position aux = new Position(m_side, ball.m_position.X, ball.m_position.Y, player.m_position.X, player.m_position.Y, 0);
	
			if (aux.dist < d) 
			{
				local = player;
				d = aux.dist;
			}
		}

		return local;
	}

	public PlayerInfo TheirGoalie()
	{
		PlayerInfo goalie = null;
		float x = -9999;		// in theory, the x of the goalie

		for (int c=0; c < m_memory.o_players.size(); c++)
		{
			PlayerInfo player = (PlayerInfo)m_memory.o_players.elementAt(c);

			if (player.m_position.X > x) 
			{
				x=player.m_position.X;
				goalie = player;
			}
		}

		return goalie;
	}

	public PlayerInfo TheirLastDefensor()
	{
		PlayerInfo last = null;
		float x = -9999;		// in theory, the x of the last defensor

		PlayerInfo goalie = TheirGoalie();

		if (goalie != null)
		{
			for (int c=0; c < m_memory.o_players.size(); c++)
			{
				PlayerInfo player = (PlayerInfo)m_memory.o_players.elementAt(c);

				if (player.m_number != goalie.m_number)
				{
					if (player.m_position.X > x) 
					{
						x=player.m_position.X;
						last = player;
					}
				}
			}
		}

		return last;
	}
}
