import java.lang.Math;

class Fuzzy
{
	Memory m_memory;
	char m_side;
	int m_number;
	String m_team;

	float KickGoalDir;
	float PassDir;

	Functions f = new Functions();

	public Fuzzy(Memory memory, char side, int number, String team)
	{
		m_memory = memory;
		m_side = side;
		m_number = number;
		m_team = team;
		KickGoalDir = 0;
		PassDir = 0;
	}

	public float Centered(float value, float min, float change1, float change2, float max)
	{
		if (value <= min) return 0;
		else if ((value <= change1) && (value > min)) return (value/(change1-min)) + (1-(change1/(change1-min)));
		else if ((value <= change2) && (value > change1)) return 1;
		else if ((value <= max) && (value > change2)) return -(value/(max-change2)) + (1+(change2/(max-change2)));
		else return 0;
	}
	public float Far(float value, float min, float max)
	{
		if (value <= min) return 0;
		else if ((value < max) && (value > min)) return (value/(max-min)) + (1-(max/(max-min)));
		else return 1;
	}

	public float Near(float value, float min, float max)
	{
		if (value <= min) return 1;
		else if ((value < max) && (value > min)) return -(value/(max-min)) + (1+(min/(max-min)));
		else return 0;
	}

	public float GoodPrediction(BallInfo ball)
	{
		return Near(ball.m_prediction, 0, SoccerParams.BALL_PREDICTION_CONFIDENCE);
	}

	public float GoodPrediction(PlayerInfo player)
	{
		return Near(player.m_prediction, 0, SoccerParams.PLAYER_PREDICTION_CONFIDENCE);
	}

	public float GoodDirection(BallInfo ball)
	{
		float dir = (float)Math.abs(ball.m_direction);

		float lim = 20/ball.m_distance;
		if (lim < 4) lim = 4;
		else if (lim > 20) lim = 20;

		return Near(dir, lim-lim/2, lim+lim/2);
	}

	public float CanKick(BallInfo ball)
	{
		float min = (float)(SoccerParams.KICKABLE_AREA - SoccerParams.KICKABLE_AREA/10);
		float max = (float)(SoccerParams.KICKABLE_AREA + SoccerParams.KICKABLE_AREA/10);

		return Near(ball.m_distance, min, max);
	}

	public float InsidePenaltyArea(float x, float y)
	{
		float condition1 = Centered(x, -53, -52, -37, -36);
		float condition2 = Centered(y, -21, -19, 19, 21);

		return f.and(condition1, condition2);
	}

	public float CachableBall(BallInfo ball)
	{
		float l = (float)Math.abs(ball.m_distance*Math.cos(f.radiants(ball.m_direction)));
		float lmin = (float)(SoccerParams.CATCHABLE_AREA_L - SoccerParams.CATCHABLE_AREA_L/10);
		float lmax = (float)(SoccerParams.CATCHABLE_AREA_L + SoccerParams.CATCHABLE_AREA_L/10);

		float condition1 = Near(l, lmin, lmax);

		float w = (float)Math.abs(ball.m_distance*Math.sin(f.radiants(ball.m_direction)));
		float wmin = (float)(SoccerParams.CATCHABLE_AREA_W - SoccerParams.CATCHABLE_AREA_W/10);
		float wmax = (float)(SoccerParams.CATCHABLE_AREA_W + SoccerParams.CATCHABLE_AREA_W/10);

		float condition2 = Near(w, wmin, wmax);

		return f.and(condition1, condition2);
	}

	public float CanCatch(BallInfo ball, float myX, float myY, int lastcatch, int time)
	{
		if (time-lastcatch >= SoccerParams.CATCH_BAN_CICLE)
		{
			float condition1 = InsidePenaltyArea(ball.m_position.X, ball.m_position.Y);
			float condition2 = InsidePenaltyArea(myX, myY);
			float condition3 = CachableBall(ball);

			return f.and(f.and(condition1, condition2), condition3);
		}

		return 0;
	}

	public float KickLong(float dir, int power, BallInfo ball)
	{
		BallInfo future_ball = new BallInfo(ball);

		float ball_vel = (float)Math.sqrt(future_ball.m_velX*future_ball.m_velX+future_ball.m_velY*future_ball.m_velY);
		float dist = power*ball.KickRate() + ball_vel;
		ball_vel *= SoccerParams.BALL_DECAY;

		while (ball_vel > 0.4)
		{
			dist += ball_vel;
			ball_vel *= SoccerParams.BALL_DECAY;
		}

		return dist;
	}


	public float AccessibleKick(float dir, float dist, BallInfo ball)
	{
		float k = KickLong(dir, SoccerParams.MAX_POWER, ball);

		return Near(dist, k-k/5, k+k/5);
	}

	public float KickGoal(BallInfo ball)
	{
		float condition1;
		float condition2;
		float kick = 0;
		float bestkick = 0;
		Position aux = null;

		float c = 0; 
		while ((c <= SoccerParams.GOAL_WIDTH/2)&&(bestkick != 1))
		{
			condition1 = 0;
			condition2 = 0;

			if (ball.m_position.Y > 0) aux = new Position(m_side, (float)(SoccerParams.FIELD_WIDTH/2), c, m_memory.myX, m_memory.myY, m_memory.myDir);
			else aux = new Position(m_side, (float)(SoccerParams.FIELD_WIDTH/2), -c, m_memory.myX, m_memory.myY, m_memory.myDir);

			condition1 = AccessibleKick(aux.dir, aux.dist, ball);
			condition2 = f.not(BusySector(aux.dir, aux.dist, SoccerParams.KICK_VIEW_MOMENT));

			kick = f.and(condition1, condition2);

			if (kick > bestkick)
			{
				bestkick = kick;
				KickGoalDir = aux.dir;
			}

			if (ball.m_position.Y > 0) aux = new Position(m_side, (float)(SoccerParams.FIELD_WIDTH/2), -c, m_memory.myX, m_memory.myY, m_memory.myDir);
			else aux = new Position(m_side, (float)(SoccerParams.FIELD_WIDTH/2), c, m_memory.myX, m_memory.myY, m_memory.myDir);

			condition1 = AccessibleKick(aux.dir, aux.dist, ball);
			condition2 = f.not(BusySector(aux.dir, aux.dist, SoccerParams.KICK_VIEW_MOMENT));

			kick = f.and(condition1, condition2);

			if (kick > bestkick)
			{
				bestkick = kick;
				KickGoalDir = aux.dir;
			}

			c+= (float)0.5;
		}

		return kick;
	}

	public float BusySector(float dir, float dist, float moment)
	{
		float busy = 0;
		float condition1;
		float condition2;

		for (int c=0; c < m_memory.o_players.size(); c++)
		{
			PlayerInfo player = (PlayerInfo)m_memory.o_players.elementAt(c);

			float d = (float)Math.abs(player.m_direction);

			condition1 = Near(d, moment/2-moment/10, moment/2+moment/10);
			condition2 = Near(player.m_distance, dist-dist/5, dist+dist/5);

			busy = f.or(busy, f.and(condition1, condition2));
		}

		return busy;
	}

	public float BusyZone(float dir, float dist, float height)
	{
		float busy = 0;
		float condition1;
		float condition2;

		for (int c=0; c < m_memory.o_players.size(); c++)
		{
			PlayerInfo player = (PlayerInfo)m_memory.o_players.elementAt(c);

			float h = (float)Math.abs(player.m_distance*Math.sin(f.radiants(player.m_direction)));

			condition1 = Near(h, height/2-height/10, height/2+height/10);
			condition2 = Near(player.m_distance, dist-dist/5, dist+dist/5);

			busy = f.or(busy, f.and(condition1, condition2));
		}

		return busy;
	}

	public float GoodPass(PlayerInfo player)
	{
		float condition1 = Far(player.m_position.X - m_memory.myX, -2, 2);
		float condition2 = Far(player.m_distance, 0, 6);

		return f.and(condition1, condition2);
	}

	public float CanPass(BallInfo ball)
	{
		float pass;
		float bestpass = -1;
		float condition1 = 0;
		float condition2 = 0;
		float condition3 = 0;
		float condition4 = 0;
		PlayerInfo player=null;

		for (int priority = 1; priority <= 6; priority++)
		{
			player = m_memory.getPlayer(m_team, f.PassPriority(m_number, priority));

			if (player != null)
			{
				condition1 = GoodPrediction(player);
				condition2 = AccessibleKick(player.m_direction, player.m_distance, ball);
				condition3 = BusySector(player.m_direction, player.m_distance-1, SoccerParams.PASS_VIEW_MOMENT);
				condition4 = GoodPass(player);

				pass = f.and(condition1, f.and(condition2, f.and(f.not(condition3), condition4)));

				if (pass > bestpass) 
				{
					bestpass = pass;
					PassDir = player.m_direction;
				}
			}
		}

		return bestpass;
	}

	public float InfluenceArea(int m_rol, float x, float y)
	{
		float condition1 = 0;
		float condition2 = 0;
		float INFINITE = 60;
		float XPENALTY = (float)(SoccerParams.FIELD_WIDTH/2-SoccerParams.PENALTY_WIDTH);
		float XMARK = (float)(SoccerParams.FIELD_WIDTH/6);
		float XMARK2 = (float)(SoccerParams.FIELD_WIDTH/10);
		float ZERO = 0;

		if (m_rol == Rols.PORTER) 
			condition1 = Centered(x, -INFINITE, -INFINITE, -XPENALTY, -XPENALTY+10);
		else if (m_rol == Rols.CENTRAL)
			condition1 = Centered(x, -INFINITE, -INFINITE, -XMARK, ZERO);
		else if ((m_rol == Rols.LATERAL_DRET)
		       ||(m_rol == Rols.LATERAL_ESQUERRA)
		       ||(m_rol == Rols.PIVOT))
			condition1 = Centered(x, -INFINITE, -INFINITE, ZERO, XMARK2);
		else if ((m_rol == Rols.MIG_DRET)
		       ||(m_rol == Rols.MIG_ESQUERRA)) 
			condition1 = Centered(x, -XPENALTY, -XMARK, XPENALTY, INFINITE);
		else if ((m_rol == Rols.EXTREM_DRET)
		       ||(m_rol == Rols.EXTREM_ESQUERRA)
		       ||(m_rol == Rols.MITJA_PUNTA))
			condition1 = Centered(x, -XMARK2, ZERO, INFINITE, INFINITE);
		else if (m_rol == Rols.DAVANTER_CENTRE)
			condition1 = Centered(x, -XMARK2, ZERO, INFINITE, INFINITE);

		float YGOAL = (float)(SoccerParams.GOAL_WIDTH/2);
		float YPENALTY = (float)(SoccerParams.PENALTY_HEIGHT/2);

		if (m_rol == Rols.PORTER)
			condition2 = Centered(y, -4*YPENALTY/3, -YPENALTY, YPENALTY, 4*YPENALTY/3);
		else if ((m_rol == Rols.PIVOT)
		       ||(m_rol == Rols.MITJA_PUNTA))
			condition2 = Centered(y, -INFINITE, -INFINITE, INFINITE, INFINITE);
		else if ((m_rol == Rols.CENTRAL)
		  ||(m_rol == Rols.DAVANTER_CENTRE)) 
			condition2 = Centered(y, -INFINITE, -YGOAL, YGOAL, INFINITE);
		else if ((m_rol == Rols.LATERAL_ESQUERRA)
		  ||(m_rol == Rols.MIG_ESQUERRA)
		  ||(m_rol == Rols.EXTREM_ESQUERRA)) 
			condition2 = Centered(y, -INFINITE, -INFINITE, -YGOAL, ZERO);
		else if ((m_rol == Rols.LATERAL_DRET)
		  ||(m_rol == Rols.MIG_DRET)
		  ||(m_rol == Rols.EXTREM_DRET)) 
			condition2 = Centered(y, ZERO, YGOAL, INFINITE, INFINITE);

		return f.and(condition1, condition2);
	}

	public float Necessity(int m_rol, float x, float y)
	{
		float condition1 = 0;
		float condition2 = 0;

		if (m_rol == Rols.PORTER) 
			condition1 = Centered(x, -53, -52, -46, -40);
		else if ((m_rol == Rols.LATERAL_DRET)
		       ||(m_rol == Rols.LATERAL_ESQUERRA)) 
			condition1 = Centered(x, -53, -37, -35, -15);
		else if (m_rol == Rols.CENTRAL)
			condition1 = Centered(x, -42, -37, -35, -30);
		else if (m_rol == Rols.PIVOT)
			condition1 = Centered(x, -30, -23, -21, -10);
		else if ((m_rol == Rols.MIG_DRET)
		       ||(m_rol == Rols.MIG_ESQUERRA)) 
			condition1 = Centered(x, -15, -4, -2, 15);
		else if (m_rol == Rols.MITJA_PUNTA)
			condition1 = Centered(x, -10, -1, 1, 15);
		else if ((m_rol == Rols.EXTREM_DRET)
		       ||(m_rol == Rols.DAVANTER_CENTRE) 
		       ||(m_rol == Rols.EXTREM_ESQUERRA)) 
			condition1 = Centered(x, 15, 35, 37, 53);

		if ((m_rol == Rols.LATERAL_ESQUERRA)
		  ||(m_rol == Rols.MIG_ESQUERRA)
		  ||(m_rol == Rols.EXTREM_ESQUERRA))
			condition2 = Centered(y, -35, -21, -19, -14);
		else if  ((m_rol == Rols.LATERAL_DRET)
			||(m_rol == Rols.MIG_DRET)
			||(m_rol == Rols.EXTREM_DRET))
			condition2 = Centered(y, 14, 19, 21, 35);
		else if ((m_rol == Rols.PORTER)
		       ||(m_rol == Rols.PIVOT)
		       ||(m_rol == Rols.DAVANTER_CENTRE)
		       ||(m_rol == Rols.CENTRAL)
		       ||(m_rol == Rols.MITJA_PUNTA))
			condition2 = Centered(y, -14, -1, 1, 14);

		return f.and(condition1, condition2);
	}

	public float FreeKickZone(int m_rol, float x, float y)
	{
		float condition1 = 0;
		float condition2 = 0;

		if (m_rol == Rols.PORTER) 
			condition1 = Centered(x, -54, -53, -36, -35);
		else if ((m_rol == Rols.LATERAL_DRET)
		       ||(m_rol == Rols.LATERAL_ESQUERRA)) 
			condition1 = Centered(x, -54, -53, 0, 1);
		else if (m_rol == Rols.CENTRAL)
			condition1 = Centered(x, -37, -36, -10, -9);
		else if (m_rol == Rols.PIVOT)
			condition1 = Centered(x, -11, -10, 10, 11);
		else if ((m_rol == Rols.MIG_DRET)
		       ||(m_rol == Rols.MIG_ESQUERRA)) 
			condition1 = Centered(x, -1, 0, 36, 37);
		else if (m_rol == Rols.MITJA_PUNTA)
			condition1 = Centered(x, 9, 10, 36, 37);
		else if ((m_rol == Rols.EXTREM_DRET)
		       ||(m_rol == Rols.DAVANTER_CENTRE) 
		       ||(m_rol == Rols.EXTREM_ESQUERRA)) 
			condition1 = Centered(x, 35, 36, 53, 54);

		if ((m_rol == Rols.LATERAL_ESQUERRA)
		  ||(m_rol == Rols.MIG_ESQUERRA)
		  ||(m_rol == Rols.EXTREM_ESQUERRA))
			condition2 = Centered(y, -36, -35, -17, -16);
		else if  ((m_rol == Rols.LATERAL_DRET)
			||(m_rol == Rols.MIG_DRET)
			||(m_rol == Rols.EXTREM_DRET))
			condition2 = Centered(y, 16, 17, 35, 36);
		else if ((m_rol == Rols.PORTER)
		       ||(m_rol == Rols.PIVOT)
		       ||(m_rol == Rols.DAVANTER_CENTRE)
		       ||(m_rol == Rols.CENTRAL)
		       ||(m_rol == Rols.MITJA_PUNTA))
			condition2 = Centered(y, -18, -17, 17, 18);

		return f.and(condition1, condition2);
	}

	public float BallNear(float x, float y, BallInfo ball, float min, float max)
	{
		Position aux = new Position('l', ball.m_position.X, ball.m_position.Y, x, y, 0);
		float d = aux.dist;

		return Near(d, min, max);
	}

	public float InsideZone(float d, float min, float max)
	{
		return Near(d, min, max);
	}

	public float Zone(int rol, float x, float y, BallInfo ball)
	{
		float x2 = f.XZone(rol, ball);
		float y2 = f.YZone(rol, ball);

		float d = 100;
		if ((x2 != -1)&&(y2 != -1)) d = (float)Math.sqrt((x-x2)*(x-x2) + (y-y2)*(y-y2));

		return InsideZone(d, 3, 6);
	}

	public float CenterZone(float x, float y)
	{
		float INFINITE = 99;
		float XMARK1 = (float)(11*SoccerParams.FIELD_WIDTH/35);
		float XMARK2 = (float)(13*SoccerParams.FIELD_WIDTH/35);
		float YPENALTY = (float)(SoccerParams.PENALTY_HEIGHT/2-5);

		float condition1 = Centered(x, XMARK1, XMARK2, INFINITE, INFINITE);
		float condition2 = Centered(y, 0, YPENALTY, INFINITE, INFINITE);
		float condition3 = Centered(y, -INFINITE, -INFINITE, -YPENALTY, 0);

		float zone1 = f.and(condition1, condition2);
		float zone2 = f.and(condition1, condition3);

		return f.or(zone1, zone2);
	}
}